#!/usr/bin/env python

"""
HTML linking scheme.

Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.links.common import Linker, resolve
from urllib import quote, quote_plus
from urlparse import urlparse

class HTMLLinker(Linker):

    "Translate Moin links into HTML links."

    name = "html"

    def get_top_level(self):

        "Return a relative link to the top level."

        # The root page is at the top level already.

        pagename = self.metadata.get("pagename", "")

        if pagename == self.root_pagename:
            return ""

        # Siblings of the root page are actually one level below.

        levels = pagename.count("/") + 1
        return "/".join([".."] * levels)

    def is_url(self, target):

        "Return whether the 'target' references a URL."

        scheme, host, path, params, query, fragment = urlparse(target)
        return scheme and target or None

    def normalise(self, path):

        "Return a normalised form of 'path'."

        return not path.endswith("/") and "%s/" % path or path

    def translate(self, target):

        """
        Translate the 'target', returning a tuple containing the rewritten
        target string and a suitable default label.
        """

        target = target.rstrip("/")

        # Fragments. Remove the leading hash for the label.

        if target.startswith("#"):
            return self.quote(target), target.lstrip("#")

        # Sub-pages. Remove the leading slash for the label.

        if target.startswith("/"):
            return self.translate_pagename(target), target.lstrip("/")

        # Sibling (of ancestor) pages.

        if target.startswith("../"):
            return self.translate_pagename(target), None

        # Attachment or interwiki link.

        rewritten = self.translate_qualified_link(target)
        if rewritten:
            return rewritten # includes label

        # Plain URL.

        rewritten = self.is_url(target)
        if rewritten:
            return rewritten, None

        # Top-level pages.

        return self.translate_pagename(target), None

    def translate_pagename(self, target):

        "Translate the pagename in 'target'."

        # Obtain the target pagename and the fragment.
        # Split the pagename into path components.

        t = target.split("#", 1)
        p = t[0].rstrip("/").split("/")

        # Determine the actual pagename referenced.
        # Replace the root pagename if it appears.

        pagename = self.metadata.get("pagename", "")
        resolved = resolve(t[0], pagename, self.root_pagename)

        # Rewrite the target using a relative link to the top level and then the
        # resolved pagename.

        top_level = self.get_top_level()
        t[0] = "%s%s" % (top_level and "%s/" % top_level or "", resolved)

        return self.quote("#".join(t))

    def translate_qualified_link(self, target):

        """
        Translate a possible qualified link 'target', returning a tuple
        containing a rewritten target and a suitable default label.

        Return None if the link is not suitable.
        """

        t = target.split(":", 1)
        if len(t) != 2:
            return None

        prefix, target = t

        # Attachment links.

        if prefix == "attachment":
            return self.translate_attachment(target), target

        # Interwiki links.

        url = self.mapping.get(prefix)
        if url:
            return self.translate_interwiki(url, target), target

        return None

    # Specific link translators.

    def translate_attachment(self, target):

        "Return a translation of the given attachment 'target'."

        return self.quote("./attachments/%s" % target)

    def translate_interwiki(self, url, target):

        "Return a translation of the given interwiki 'target'."

        return "%s%s" % (self.normalise(url), self.quote(target))

    # Path encoding.

    def quote(self, s):

        """
        Quote URL path 's', preserving path separators and fragment indicators,
        encoding fragment identifiers.
        """

        parts = s.split("#", 1)

        if len(parts) > 1:
            parts[1] = self.make_id(parts[1])

        return "#".join(map(quote, parts))

    # Identifier encoding.

    def make_id(self, s):

        "Make a suitable identifier for HTML element identification."

        # NOTE: This reproduces the Moin algorithm for compatibility.
        # NOTE: There may well be improvements possible, possibly by replacing plus
        # NOTE: with something less cumbersome, even though plus may be unusual in
        # NOTE: things like headings, anyway.

        # The desired output is the following pattern:

        # [A-Za-z][-_:.A-Za-z0-9]*

        # The Python UTF-7 encoder preserves symbols and it encodes + as +- with an
        # output range as follows (in addition to A-Za-z0-9):

        # -_:.%+ !"#$&\'()*,/;<=>?@[]^`{|}

        # The quote_plus function converts space to plus, preserves -_:. and encodes
        # all other symbols (including original occurrences of plus and percent) and
        # non-alphanumeric (ASCII) characters using percent encoding.

        # With colons preserved, the resulting output is in the following range
        # (in addition to A-Za-z0-9):

        # -_:.%+

        # Percent will only occur as an encoding prefix. Plus will only occur as a
        # replacement for space.

        # Combining quote_plus and UTF-7 gives the following range (in addition to
        # A-Za-z0-9):

        # -_:.%+

        # Examples:

        #          UTF-7         quote_plus    replace percent and plus
        # :     -> :          -> :          -> :
        # -     -> -          -> -          -> -
        # .     -> .          -> .          -> .
        # %     -> %          -> %25        -> .25
        # +     -> +-         -> %2B-       -> .2B-
        # _     -> _          -> _          -> _
        # space -> space      -> +          -> _

        # See: RFC2152 - UTF-7 A Mail-Safe Transformation Format of Unicode

        quoted = quote_plus(s.encode("utf-7"), ":").replace("%", ".").replace("+", "_")

        # Ensure that the identifier starts with an alphabetical character.

        if not quoted[0].isalpha():
            return "A%s" % quoted
        else:
            return quoted

linker = HTMLLinker

# vim: tabstop=4 expandtab shiftwidth=4
